import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
import seaborn as sns

def plot_crosstab_barplot(data,y=None,by=None,figsize = (12,6),ylab=float('nan')
                          ,title=None,loc='best',anchor=(1,1)):
    """
    data: Dataframe to plot
    y: Column name to crosstabulate by 
	by: Column name for which to plot the percentage breakdown 
	figsize: (10,6). Overall figure size for the Violin and Box plots 
    ylab: Label to use for the Y-axis
    title: (Default = None) Title of the legend
    loc: (Default = 'best') Legend location
    anchor: Default = (1,1). Specfies where to anchor the legend
    """
    crosstab_data = pd.crosstab(data[by],data[y].astype(bool))
    crosstab_rate = round(crosstab_data.div(crosstab_data.sum(axis = 1), axis = 0),2)

    ax = crosstab_rate.plot(kind='barh'
                            ,stacked=True
                            ,figsize = figsize
                            #,colormap="PiYG"
                           )
    if (pd.notna(ylab)):
        ax.set_ylabel(ylab)
        
    plt.legend(title=title, loc=loc, fancybox=True, bbox_to_anchor=anchor)

       


from .plot_learning_curves import plot_learning_curves, plot_learning_curve
from .plot_hyper_validation_curves import plot_hyper_validation_curves
from .train_classifier import train_classifier
from .plot_classification import print_classification_details, plot_classification_report



import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
import seaborn as sns
import math

from sklearn.model_selection import learning_curve
from .common import time_usage

#########################
#### Learning Curves ####
#########################


def plot_learning_curve(estimator, title_suffix, X, y, scoring="accuracy", ylim=None, cv=None,
                         n_jobs=None, train_sizes=np.linspace(.2, 1.0, 5),verbose=1, ax = None):
    """
    NOTE: Adopted from: https://scikit-learn.org/stable/auto_examples/model_selection/plot_learning_curve.html,
    but allows plotting multiple metrics at a time when called from within plot_learning_curves (note the s at the end)
    
    Generate a simple plot of the test and training learning curve.

    Parameters
    ----------
    estimator : object type that implements the "fit" and "predict" methods
        An object of that type which is cloned for each validation.

    title : string
        Title for the chart.

    X : array-like, shape (n_samples, n_features)
        Training vector, where n_samples is the number of samples and
        n_features is the number of features.

    y : array-like, shape (n_samples) or (n_samples, n_features), optional
        Target relative to X for classification or regression;
        None for unsupervised learning.

    ylim : tuple, shape (ymin, ymax), optional
        Defines minimum and maximum yvalues plotted.

    cv : int, cross-validation generator or an iterable, optional
        Determines the cross-validation splitting strategy.
        Possible inputs for cv are:
          - None, to use the default 3-fold cross-validation,
          - integer, to specify the number of folds.
          - :term:`CV splitter`,
          - An iterable yielding (train, test) splits as arrays of indices.

        For integer/None inputs, if ``y`` is binary or multiclass,
        :class:`StratifiedKFold` used. If the estimator is not a classifier
        or if ``y`` is neither binary nor multiclass, :class:`KFold` is used.

        Refer :ref:`User Guide <cross_validation>` for the various
        cross-validators that can be used here.

    n_jobs : int or None, optional (default=None)
        Number of jobs to run in parallel.
        ``None`` means 1 unless in a :obj:`joblib.parallel_backend` context.
        ``-1`` means using all processors. See :term:`Glossary <n_jobs>`
        for more details.

    train_sizes : array-like, shape (n_ticks,), dtype float or int
        Relative or absolute numbers of training examples that will be used to
        generate the learning curve. If the dtype is float, it is regarded as a
        fraction of the maximum size of the training set (that is determined
        by the selected validation method), i.e. it has to be within (0, 1].
        Otherwise it is interpreted as absolute sizes of the training sets.
        Note that for classification the number of samples usually have to
        be big enough to contain at least one sample from each class.
        (default: np.linspace(0.1, 1.0, 5))
    """
    
    
    train_sizes, train_scores, test_scores = learning_curve(estimator
                                                            ,X,y
                                                            ,cv=cv
                                                            ,scoring=scoring
                                                            ,n_jobs=n_jobs
                                                            ,train_sizes=train_sizes
                                                            ,verbose=verbose
                                                           )
    train_scores_mean = np.mean(train_scores, axis=1)
    train_scores_std = np.std(train_scores, axis=1)
    test_scores_mean = np.mean(test_scores, axis=1)
    test_scores_std = np.std(test_scores, axis=1)
    

    ax.fill_between(train_sizes, train_scores_mean - train_scores_std,
                     train_scores_mean + train_scores_std, alpha=0.1,
                     color="r")
    ax.fill_between(train_sizes, test_scores_mean - test_scores_std,
                     test_scores_mean + test_scores_std, alpha=0.1, color="g")
    ax.plot(train_sizes, train_scores_mean, 'o-', color="r",
             label="Training score")
    ax.plot(train_sizes, test_scores_mean, 'o-', color="g",
             label="Cross-validation score")

    ## Annotate Plots
    ax.title.set_text("Learning Curve: " + title_suffix)
    if (ylim == None):
        ax.set_ylim(-0.1, 1.1)
    if ylim is not None:
        ax.set_ylim(*ylim)
       
    ax.set_ylabel(title_suffix + " Score")
    ax.set_xlabel("Training examples")
    ax.grid(b=True)
    ax.legend(loc="best")
    #return plt


def plot_learning_curves(estimator,X,y,scoring,cv=None,n_jobs=None,verbose=1,arFigsize=None):
    num_rows = math.ceil(len(scoring)/2)
    if (arFigsize == None):
        arFigsize = (12,num_rows*4)
    
    fig, axes = plt.subplots(num_rows, 2, figsize=arFigsize, squeeze=False)
    i = 0
    for scorer in scoring.keys():
        with time_usage(" Learning Curve | " + scorer):
            plot_learning_curve(estimator=estimator
                                ,title_suffix  = scorer.title()
                                ,X=X, y=y
                                ,scoring=scoring[scorer]
                                ,cv=cv
                                ,n_jobs=n_jobs
                                ,verbose=verbose
                                ,ax=axes[math.floor(i/2),i%2]
                               )
            i = i+1

   #plt.show()


import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
import seaborn as sns
import warnings
import math

def plot_data(data, kind = 'dist', rows=None, cols=None, kde=False, bins=None, figsize=None, xrot=0, yrot=0):
    """
    data = Dataframe to plot
    kind = 'dist'. Specifies kind of plot, Options are 'dist' (default), 'count'
    rows = 3. Specifies number of rows in the plots
    rows = 3. Specifies number of columns in the plots
    kde = False. If kind = 'dist', should the KDE be plotted?
    bins = If kind = 'dist', this argument specifies the number of bins that should be plotted
    figsize = Overall figure size for the plot matrix. Autocalculated by default.
    xrot = 0. Degrees to rotate the X label by
    yrot = 0. Degrees to rotate the Y label by
    """

    """
    TODO: Add checks for variable types for each type of plot
    """

    if (kind == 'dist'):
        cat_columns = data.select_dtypes(include=['object','category']).columns
        data = data.select_dtypes(include=['number'])
        if (len(cat_columns) > 0):
            warnings.warn("The data has categorical columns {} for which a distribution will not be plotted".format(cat_columns))    
    elif (kind == 'count'):
        num_columns = data.select_dtypes(include=['number']).columns
        data = data.select_dtypes(include=['object','category'])
        if (len(num_columns) > 0):
            warnings.warn("The data has numeric columns {} for which a count plot will not be plotted".format(num_columns))        
    else:
        raise Exception("kind = " + str(kind) + " is not supported yet") 

    num_columns = len(data.columns)
    
    # Both are None, then code decides
    if ((rows == None) and (cols == None)):
        if (num_columns % 5 == 0):
            cols = 5
        elif (num_columns % 4 == 0):
            cols = 4
        elif (num_columns % 3 == 0):
            cols = 3
        elif (num_columns % 2 == 0):
            cols = 2
        else:
            cols = 1
        rows = math.ceil(num_columns/cols)
    # If number of rows is none, then number of columns gets decided by code 
    elif (rows == None):
        rows = math.ceil(num_columns/cols)
    # If number of columns is none, then number of rows gets decided by code
    elif (cols == None):
        cols = math.ceil(num_columns/rows)
    # When both are specified by user
    elif (rows * cols < num_columns):
        warnings.warn("Number of rows and columns specified is less than number of scatter plots to be plotted. Some scatterplots will be omitted");    
    
    num_plots_requested = rows * cols
   
    if (figsize == None):
        if (cols > 2):
            loWidth = 4 * cols
            loHeight = 4 * rows
        else:
            loWidth = 4 * cols
            loHeight = 4 * rows
        figsize = (loWidth,loHeight)
    
    fig, axes = plt.subplots(nrows=rows, ncols=cols,figsize = figsize)
    
    cols_to_plot = data.columns
          
    for i, column in enumerate(cols_to_plot):
        if i < num_plots_requested:
            if (rows == 1 and cols == 1):
                axis = axes
            elif ((rows == 1 and cols != 1) or (cols == 1 and rows != 1)):
                axis = axes[i]
            else:
                axis = axes[i//cols,i%(cols)]
            
            if (kind == 'dist'):
                ax = sns.distplot(data[data[column].notna()][column],kde=kde, bins = bins, ax=axis)
                ax.set_xticklabels(ax.get_xticklabels(), rotation = xrot) # , fontsize = 8
                ax.set_yticklabels(ax.get_yticklabels(), rotation = yrot) # , fontsize = 8
            elif (kind == 'count'):
                ax = sns.countplot(x = column, data = data, ax = axis)
                ax.set_xticklabels(ax.get_xticklabels(), rotation = xrot) # , fontsize = 8
                ax.set_yticklabels(ax.get_yticklabels(), rotation = yrot) # , fontsize = 8
            else:
                pass;
                
    
    plt.tight_layout()


import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
import seaborn as sns

from pandas.plotting import parallel_coordinates

def plot_parallel_coordinates(data, by, sample=True, frac=1.0, normalize=True
                              ,figsize = (8,6),legend_title=None
                              ,legend_loc='best',legend_anchor=(1,1),xrot = 0, random_state=1):
    """
    data: Dataframe to use for plotting the parallel coordinates plot
    by: Column name to categorize the plot by (currently only supports a column with 2 classes having values 0 and 1)
    sample: (Default = True) Should the data be sampled (may be necessary for large datasets)
    frac: (Default = 1.0 --> Take full dataset) If sample = True, then use this sampling fraction
    normalize: (Defauult = True) Should the data be normalized. Make True if columns are not on same scale already 
    """
    
    if(sample):    
        df_sub = data.sample(frac=frac,replace=False,random_state=random_state)
    else:
        df_sub = data.copy(deep=False)
    
    # Commenting out since the by variable could have more than 2 levels              
    #df_sub[by] = df_sub[by] == 1 # converting categorical variable into number for plotting

    
    # This plot is more meaningful when values are normalized
    # They are on the same scale here so no need to normalize
    if (normalize):
        df_normalized = (df_sub-df_sub.mean())/(df_sub.std())
        df_normalized[by] = df_sub[by]
    else:
        df_normalized = df_sub.copy(deep=False)

    plt.figure(figsize = figsize)
    
    ax = parallel_coordinates(df_normalized,by,colormap='viridis')
    if (legend_title == None):
        legend_title = by
    plt.legend(title=legend_title, loc=legend_loc, fancybox=True, bbox_to_anchor=legend_anchor)
    ax.set_xticklabels(ax.get_xticklabels(), rotation = xrot) # , fontsize = 8



import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
import seaborn as sns
import warnings
import math

def plot_scatter_one_vs_all(data, y = None, hue = None, rows=None, cols=None, figsize=None, xrot=0, yrot=0, alpha = 0.7):
    """
    data: (Mandatory) Dataframe to plot
    y: (Mandatory) Variable to plot on the y-axis
    hue: Used to define the column in data that is used to color code the scatterplot 
	rows: (Autocalculated by default) Specifies number of rows in the plots
    cols: (Autocalculated by default)Specifies number of columns in the plots
    figsize: (Autocalculated by default) Overall figure size for the plot matrix. Autocalculated by default.
    xrot: (Default = 0) Degrees to rotate the X label by
    yrot: (Default = 0) Degrees to rotate the Y label by
    alpha: (Default = 0.7) Transparency to use for the scatterplots
    """

    """
	TODO:
		(1) Check for datatypes to make sure they are correct
		(2) Values not showing up on axis
    """
	
    if (y == None):
        raise Exception("Please specify the variable to plot on the Y-axis");
		
    #numeric_columns = data.select_dtypes(include=['number']).columns
    #categorical_columns = data.select_dtypes(include=['object','category']).columns
    num_columns = len(data.columns)
    
    if (hue == None):
        num_scatter = num_columns - 1
    elif (hue == y): # Both Hue and Y are Same
        num_scatter = num_columns - 1
    else:
        num_scatter = num_columns - 2

    # Both are None, then code decides
    if ((rows == None) and (cols == None)):
        if (num_scatter % 5 == 0):
            cols = 5
        elif (num_scatter % 4 == 0):
            cols = 4
        elif (num_scatter % 3 == 0):
            cols = 3
        elif (num_scatter % 2 == 0):
            cols = 2
        else:
            cols = 1
        rows = math.ceil(num_scatter/cols)
    # If number of rows is none, then number of columns gets decided by code 
    elif (rows == None):
        rows = math.ceil(num_scatter/cols)
    # If number of columns is none, then number of rows gets decided by code
    elif (cols == None):
        cols = math.ceil(num_scatter/rows)
    # When both are specified by user
    elif (rows * cols < num_scatter):
        warnings.warn("Number of rows and columns specified is less than number of scatter plots to be plotted. Some scatterplots will be omitted");    
        		
    num_plots_requested = rows * cols
		
    if (figsize == None):
        if (cols > 2):
            loWidth = 4 * cols
            loHeight = 4 * rows
        else:
            loWidth = 8 * cols
            loHeight = 8 * rows
        figsize = (loWidth,loHeight)
    fig, axes = plt.subplots(nrows=rows, ncols=cols,figsize = figsize)
    
    cols_to_plot = data.columns
    
	# Remove columns that you dont want to plot
    cols_to_plot = data.columns.drop(y)
    if (hue != None):
        cols_to_plot = data.columns.drop([y,hue])
        
    for i, column in enumerate(cols_to_plot):
        if i < num_plots_requested:
            if (rows == 1 and cols == 1):
                axis = axes
            elif ((rows == 1 and cols != 1) or (cols == 1 and rows != 1)):
                axis = axes[i]
            else:
                axis = axes[i//cols,i%(cols)]
        
            ax = sns.scatterplot(data=data, y = y, x = column ,hue = hue, alpha = alpha, ax = axis)
            ax.set_xticklabels(ax.get_xticklabels(), rotation = xrot) # , fontsize = 8
            ax.set_yticklabels(ax.get_yticklabels(), rotation = yrot) # , fontsize = 8
        
    plt.tight_layout()


import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
import seaborn as sns

from sklearn.preprocessing import StandardScaler
from sklearn.decomposition import PCA

def perform_pca(data, n_components = 3, y=None):    
    """
    data = Data on which to perform PCA
    n_components = 3. This is the number of Principal Components to consider for the analysis
    """
    # Separating and standardizing the features
    features = data.copy(deep=True)
    if (y != None):
        features = data.drop(y,axis=1)
    x = StandardScaler().fit_transform(features)

    pca = PCA(n_components = n_components)
    principalComponents = pca.fit_transform(x)
    principal_comp = pd.DataFrame(data = principalComponents, 
                                  columns = ['Principal Component ' + str(i) for i in range(1,(n_components+1))])
    if (y != None):
        final_df = pd.concat([principal_comp, data[y]], axis = 1)
    else:
        final_df = principal_comp.copy(deep=True)
    print("% variance explained by each Principal Component: " + str(np.round(pca.explained_variance_ratio_ * 100,2)))
    print("% Total variance explained by all Principal Components: " + str(round(sum(pca.explained_variance_ratio_) * 100,2)) + "%")
    return (final_df)


def plot_prin_comps(data, figsize=None, hue=None, alpha = 0.7, bins=None, kde=True, xrot=0, yrot=0):
    """
    data = Data returned by the perform_pca function
    figsize = Overall figure size for the overall plot. Autocalculated by default based on number of columns
    hue = 'promotion_max'. Column name to use to color code the scatter plot
    alpha = 0.7. Transparency to use for the scatterplots
    bins = 50. Number of bins to use in the histogram on the diaginals
    kde = True. Should KDE be plotted on the diagonals
    xrot = 0
    yrot = 0
    """
    # Define Shape and size of matrix
    arSquare = 1
    if (hue != None):
        arSquare = data.shape[1]-1 # remove the column hue for size calculations
    else:
        arSquare = data.shape[1] 
		
    if (figsize == None):
        figsize = (4*arSquare,4*arSquare)
        
    fig, axes = plt.subplots(nrows=arSquare, ncols=arSquare,figsize = figsize)
    
    cols_to_plot = data.columns
    if (hue != None):
        cols_to_plot = data.columns.drop(hue)
		
    for i, column1 in enumerate(cols_to_plot):
        for j, column2 in enumerate(cols_to_plot):
            if (i == j):
                # draw histogram and KDE on diagonal   
                ax = sns.distplot(data[data[column1].notna()][column1],kde=kde,bins=bins, ax=axes[i,j])
                ax.set_xticklabels(ax.get_xticklabels(), rotation = xrot) # , fontsize = 8
                ax.set_yticklabels(ax.get_yticklabels(), rotation = yrot) # , fontsize = 8
                           
            if (i < j):
                # only draw scatterplot on one side of the diagnal
                ax = sns.scatterplot(data=data, y = column1, x = column2, hue = hue, ax = axes[i,j])
                ax.set_xticklabels(ax.get_xticklabels(), rotation = xrot) # , fontsize = 8
                ax.set_yticklabels(ax.get_yticklabels(), rotation = yrot) # , fontsize = 8
        
    plt.tight_layout()
    
    


import numpy as np
import matplotlib.pyplot as plt

# this code has been manipulated from the source available on sklearn's website documentation
import itertools
from sklearn import metrics as mt
from scipy import linalg
import matplotlib as mpl
from sklearn import mixture

class GaussianMixtureHelper:
    def __init__(self, X, y=None
                 ,n_components_range=range(2,3) ,cov_types = ['spherical']
                 ,metric = 'bic', random_state=101):
        """
        Class to train and evaluate a Gaussian Mixture Cluster Model
        """
        self.X = X
        self.y = y
        self.n_components_range = n_components_range
        self.cov_types = cov_types 
        # Add exception here is metric is not of the right type
        self.metric = metric
        self.random_state = random_state
        self.y_pred = None
        self.best_gmm = None
        self.best_gmm_bic = None
        self.best_gmm_aic = None
        self.bic = []
        self.aic = []
        self.lowest_bic = np.infty
        self.lowest_aic = np.infty
            
    def train(self):
        """
        Train the Gaissian Mixture Model across a range of cluster values and covariance types
        """
        for cov_type in self.cov_types:
            for n_components in self.n_components_range:
                # Fit a mixture of Gaussians with EM
                gmm = mixture.GaussianMixture(n_components=n_components
                                                  ,covariance_type=cov_type
                                                  ,random_state=self.random_state)
                gmm.fit(self.X)
                self.bic.append(gmm.bic(self.X))
                self.aic.append(gmm.aic(self.X))
                
                if self.bic[-1] < self.lowest_bic:
                    self.lowest_bic = self.bic[-1]
                    self.best_gmm_bic = gmm
                    
                if self.aic[-1] < self.lowest_aic:
                    self.lowest_aic = self.aic[-1]
                    self.best_gmm_aic = gmm
        
        self.set_best_model()
        self.y_pred = self.predict(self.X)
        return(self)
    
    def set_metric(self, metric):
        self.metric = metric
        
    def set_best_model(self):
        """
        Use to set the best model to the one based on a specific metric
        Default Metric = 'bic'; Other Option(s): 'aic' 
        """
        if (self.metric == 'bic'):
            self.best_gmm = self.best_gmm_bic
        elif(self.metric == 'aic'):
            self.best_gmm = self.best_gmm_aic  
    
    def get_best_model(self):
        return(self.best_gmm)
    
    def plot_metrics(self, figsize = (12,4)):
        # this code has been manipulated from the source available on sklearn's website documentation
        # plot the BIC
        
        plt.figure(figsize=figsize)
               
        # Plot the BIC scores
        spl = plt.subplot(1,2,1)
        color_iter = itertools.cycle(['k', 'r', 'b', 'g', 'c', 'm', 'y'])
        bars = []
        self.bic = np.array(self.bic)
        
        for i, (self.cov_type, color) in enumerate(zip(self.cov_types, color_iter)):
            xpos = np.array(self.n_components_range) + .2 * (i - 2)
            bars.append(plt.bar(xpos, self.bic[i * len(self.n_components_range):
                                          (i + 1) * len(self.n_components_range)],
                                width=.2, color=color))
        plt.xticks(self.n_components_range)
        plt.ylim([self.bic.min() * 1.01 - .01 * self.bic.max(), self.bic.max()])
        plt.title('BIC score per model')

        xpos = np.min(self.n_components_range)-0.4 + np.mod(self.bic.argmin(), len(self.n_components_range)) +\
            .2 * np.floor(self.bic.argmin() / len(self.n_components_range))
        plt.text(xpos, self.bic.min() * 0.97 + .03 * self.bic.max(), '*', fontsize=14)
        spl.set_xlabel('Number of components')
        spl.legend([b[0] for b in bars], self.cov_types)

        # Plot the AIC scores
        spl = plt.subplot(1,2,2)
        color_iter = itertools.cycle(['k', 'r', 'b', 'g', 'c', 'm', 'y'])
        bars = []
        self.aic = np.array(self.aic)
        
        for i, (self.cov_type, color) in enumerate(zip(self.cov_types, color_iter)):
            xpos = np.array(self.n_components_range) + .2 * (i - 2)
            bars.append(plt.bar(xpos, self.aic[i * len(self.n_components_range):
                                          (i + 1) * len(self.n_components_range)],
                                width=.2, color=color))
        plt.xticks(self.n_components_range)
        plt.ylim([self.aic.min() * 1.01 - .01 * self.aic.max(), self.aic.max()])
        plt.title('AIC score per model')

        xpos = np.min(self.n_components_range)-0.4 + np.mod(self.aic.argmin(), len(self.n_components_range)) +\
            .2 * np.floor(self.aic.argmin() / len(self.n_components_range))
        plt.text(xpos, self.aic.min() * 0.97 + .03 * self.aic.max(), '*', fontsize=14)
        spl.set_xlabel('Number of components')
        spl.legend([b[0] for b in bars], self.cov_types)

        plt.tight_layout()
        #plt.show()
        return(plt)
    
    def predict(self,X):
        clf = self.get_best_model()
        y_pred = clf.predict(X)
        return(y_pred)
    
    def plot_best_model(self,feat_x,feat_y):
        plt.figure(figsize=(12,6))
        splot = plt.subplot(1,1,1)
        
        color_iter = itertools.cycle(['k', 'r', 'b', 'g', 'c', 'm', 'y'])
        clf = self.get_best_model()
        
        for i, (mean, covar, color) in enumerate(zip(clf.means_, clf.covariances_,color_iter)):
            if len(covar.shape)<2:
                tmp = np.zeros((2,2))
                np.fill_diagonal(tmp,covar)
                covar = tmp
            elif covar.shape[0] != covar.shape[1]:
                covar = np.diag(covar)

            v, w = linalg.eigh(covar)
            if not np.any(self.y_pred == i):
                continue
            
            plt.scatter(self.X[self.y_pred == i][feat_x], self.X[self.y_pred == i][feat_y], 5, color=color)

            # Plot an ellipse to show the Gaussian component
            angle = np.arctan2(w[0][1], w[0][0])
            angle = 180 * angle / np.pi  # convert to degrees
            v *= 4
            ell = mpl.patches.Ellipse(mean, v[0], v[1], 180 + angle, color=color)
            ell.set_clip_box(splot.bbox)
            ell.set_alpha(.5)
            splot.add_artist(ell)

        plt.title('Selected GMM')
        plt.show()
        
    def clusters_vs_true_labels(self):
        self.y_pred = self.predict(self.X)
        num_true_classes = len(set(self.y))
        print(mt.confusion_matrix(self.y,self.y_pred)[0:num_true_classes,:])
        



import warnings
import pandas as pd
from sklearn.preprocessing import StandardScaler
from sklearn import metrics
from more import viz_helper as vh

class BaseClusterWithN:
    def __init__(self, X, n_clusters, scaled = True, random_state = 101):
        """
        Class to train and evaluate a Base Cluster Class with Number of Clusters Specified
        """
        self.X = X.reset_index(drop=True)
        self.n_clusters = n_clusters
        self.scaled = scaled
        self.random_state = random_state
        self.cluster_obj = None # Define in child class
        self.labels = None
        self.silhoutte_score = None;
        self.merged_data = None
        self.merged_scaled_data = None
        self.columns = self.X.columns
        
        std_scl = StandardScaler()
        self.X_scaled = pd.DataFrame(std_scl.fit_transform(self.X), columns=self.columns)
                    
    def train(self, merge = True):
        """
        Train the clustering method
        """
        if (self.scaled):
            self.cluster_obj.fit(self.X_scaled)
        else:
            self.cluster_obj.fit(self.X)
            
        self.labels = self.cluster_obj.labels_
        
        if (merge):
            self.merge_data_labels()
            
        return(self) # Allows to cascade methods
            
    def evaluate(self, metric = "silhoutte"):
        """
        Provides the Goodness of Fit Statistics for the clustering algorithm
        """
        if (metric == "silhoutte"):
            if (self.scaled):
                self.silhoutte_score = metrics.silhouette_score(self.X_scaled, self.labels, random_state= self.random_state)
            else:
                self.silhoutte_score = metrics.silhouette_score(self.X, self.labels, random_state= self.random_state)
        else:
            warnings.warn("Metrix {} is not supported".format(metric))

        print("Silhouette Coefficient: {}".format(self.silhoutte_score))
        
    def merge_data_labels(self):
        self.merged_data = pd.concat([self.X,pd.Series(self.labels,name='labels')], axis = 1)
        self.merged_scaled_data = pd.concat([self.X_scaled,pd.Series(self.labels,name='labels')], axis = 1)
        
    def cluster_obs_count(self):
        """
        Gives the number of observations in each cluster
        """
        return(self.merged_data.groupby('labels').count().transpose().iloc[0,:])
        
    def cluster_means(self):
        """
        Provides the means of the cluster features for each cluster
        """
        return(self.merged_data.groupby('labels').mean().transpose())
        
    def cluster_means_scaled(self):
        """
        Provides the means (scaled) of the cluster features for each cluster
        """
        return(self.merged_data.groupby('labels').mean().transpose())
        
    def plot_parallel_coordinates(self, scaled=True, frac=0.05, figsize=(12,6), xrot=0):
        """
        Plot the parallel coordinate plots for the features in each cluster
        """
        if (scaled):
            vh.plot_parallel_coordinates(data = self.merged_scaled_data, by = 'labels', normalize=False, frac=frac, figsize=figsize, xrot=xrot)
        else:
            vh.plot_parallel_coordinates(data = self.merged_data, by = 'labels', normalize=False, frac=frac, figsize=figsize, xrot=xrot)
            
        
    
        
        
            
        
        
    
        
            



import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
import seaborn as sns
   
def plot_corr(data, figsize= None, xrot=0, yrot=0):
    """
    corr_mat: Correlation matrix to visualize
    figsize: Overall figure size for the overall plot
    xrot: (Default = 0) Rotate x-labels
    yrot: (Default = 0) Rotate y-labels
    """
    fig, axis = plt.subplots(1,1,figsize=figsize)
    corr_mat = data.corr()
    sns.heatmap(round(corr_mat,2),annot=True,cmap='RdBu',vmin=-1, vmax=1)
    axis.set_xticklabels(axis.get_xticklabels(), rotation = xrot)
    axis.set_yticklabels(axis.get_yticklabels(), rotation = yrot)



import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
import seaborn as sns

def plot_nulls(data, figsize=None):
    """
	data: Dataframe to plot
    """
    
    # Run heatmap again to validate nulls are handled
    plt.figure(figsize=figsize)
    sns.heatmap(data.isnull(),yticklabels=False,cbar=False,cmap='viridis')


import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
import seaborn as sns

def plot_violin_box(data, x, y, bw=1, cut=0, scale='area', figsize=(10,5) 
                    ,xrot=0, yrot=0, xlab=float('nan'), ylab=float('nan')):
    """
	A functions to visualize a continuous field vs. a categorical field)
    data: Dataframe to plot
    x: Column name to plot on the X-axis (categorical variable)
    y: Column name to plot on the Y-axis (continuous variable)
    bw: (Default = 1) Bandwidth for the Violon plot
    cut = (Default = 0) Cut option for the Violin plot
    scale: (Default = Area) Scale option for the Violin plot
    figsize: Default = (10,5). Overall figure size for the Violin and Box plots 
    xrot: (Default = 0). Degrees to rotate the X label by
    yrot: (Default = 0). Degrees to rotate the Y label by
    xlab: Label to use for the X-axis
    ylab: Label to use for the Y-axis
    """
    fig, axes = plt.subplots(1, 2, figsize=figsize)
    plt.xticks(rotation=xrot)
    
    # Violin Plot
    ax = sns.violinplot(x=x,y=y,data=data
                        ,palette="RdBu"
                        ,cut = 0
                        ,ax = axes[0],scale=scale
                        ,inner='quartile',bw=bw)
    ax.set_xticklabels(ax.get_xticklabels(), rotation = xrot) # , fontsize = 8
    ax.set_yticklabels(ax.get_yticklabels(), rotation = yrot) # , fontsize = 8
    if (pd.notna(xlab)):
        ax.set_xlabel(xlab)
    if (pd.notna(ylab)):
        ax.set_ylabel(ylab)
    
    # Box Plot
    ax = sns.boxplot(x=x,y=y,data=data,palette="RdBu",ax = axes[1])
    ax.set_xticklabels(ax.get_xticklabels(), rotation = xrot) # , fontsize = 8
    ax.set_yticklabels(ax.get_yticklabels(), rotation = yrot) # , fontsize = 8
    if (pd.notna(xlab)):
        ax.set_xlabel(xlab)
    if (pd.notna(ylab)):
        ax.set_ylabel(ylab)
        


import numpy as np
import pandas as pd
import matplotlib.pyplot as plt

from .common import time_usage
from .plot_classification import print_classification_details, plot_classification_report

def train_classifier(estimator,X,y,cv,verbose=False):
    """
    # Manual training of Classifiers using loops (does not parallelize)
    # Not recommended for detailed analysis (use Scikit Learn's internal functions with njobs for parallelism)
    # Only useful whe you want to vizualize the confusion matrix and the classification report
    """
    
    # For ROC
    printROC=False # ROC is not available to end users for now.
    tprs = []
    aucs = []
    mean_fpr = np.linspace(0, 1, 100)
    
    
    for iter_num, (train_indices, test_indices) in enumerate(cv.split(X,y)):
        with time_usage(" Classifier: Iteration " + str(iter_num+1)):
            estimator.fit(X[train_indices],y[train_indices])  # train object
            y_hat = estimator.predict(X[test_indices]) # get test set precitions
                                    
            if (printROC == True):
                print_classification_details(actual=y[test_indices], predicted=y_hat,verbose=False)
            else:
                print_classification_details(actual=y[test_indices], predicted=y_hat,verbose=verbose)
            
            if (printROC == True):    
                ## For ROC
                probas_ = estimator.predict_proba(X[test_indices])
                # Compute ROC curve and area the curve
                fpr, tpr, thresholds = mt.roc_curve(y[test_indices], probas_[:, 1])
                tprs.append(np.interp(mean_fpr, fpr, tpr))
                tprs[-1][0] = 0.0
                roc_auc = mt.auc(fpr, tpr)
                aucs.append(roc_auc)
                plt.plot(fpr, tpr, lw=1, alpha=0.3, label='ROC fold %d (AUC = %0.2f)' % (iter_num, roc_auc))
            
   
    if (printROC == True):
        plt.plot([0, 1], [0, 1], linestyle='--', lw=2, color='r',label='Chance', alpha=.8)

        mean_tpr = np.mean(tprs, axis=0)
        mean_tpr[-1] = 1.0
        mean_auc = mt.auc(mean_fpr, mean_tpr)
        std_auc = np.std(aucs)
        plt.plot(mean_fpr, mean_tpr, color='b',
                 label=r'Mean ROC (AUC = %0.2f $\pm$ %0.2f)' % (mean_auc, std_auc),
                 lw=2, alpha=.8)

        std_tpr = np.std(tprs, axis=0)
        tprs_upper = np.minimum(mean_tpr + std_tpr, 1)
        tprs_lower = np.maximum(mean_tpr - std_tpr, 0)
        plt.fill_between(mean_fpr, tprs_lower, tprs_upper, color='grey', alpha=.2,
                         label=r'$\pm$ 1 std. dev.')

        plt.xlim([-0.05, 1.05])
        plt.ylim([-0.05, 1.05])
        plt.xlabel('False Positive Rate')
        plt.ylabel('True Positive Rate')
        plt.title('ROC curve')
        plt.legend(loc="lower right")
        plt.show()


# For Time Logging
import time
from contextlib import contextmanager
import logging

@contextmanager
# Timing Function
def time_usage(name=""):
    """
    log the time usage in a code block
    """
    #print ("In time_usage runID = {}".format(runID))
    start = time.time()
    yield
    end = time.time()
    elapsed_seconds = float("%.10f" % (end - start))
    logging.info('%s: Time Taken (seconds): %s', name, elapsed_seconds)



from .plot_violin_box import plot_violin_box
from .plot_scatter_one_vs_all import plot_scatter_one_vs_all
from .plot_crosstab_barplot import plot_crosstab_barplot
from .pca_helper import perform_pca, plot_prin_comps
from .plot_corr import plot_corr
from .plot_nulls import plot_nulls
from .plot_data import plot_data
from .plot_parallel_coordinates import plot_parallel_coordinates
from .plot_similarity import plot_similarity



from sklearn.cluster import KMeans
from . import BaseClusterWithN

class KMeansHelper(BaseClusterWithN):
    def __init__(self, X, n_clusters, init = "k-means++", n_jobs = None, scaled = True, random_state = 101):
        """
        Class to train and evaluate a KMeans Cluster Model
        """
        super().__init__(X=X, n_clusters=n_clusters, random_state = random_state)
        self.init = init
        self.n_jobs = n_jobs
        self.cluster_obj = KMeans(n_clusters=self.n_clusters, init=self.init, random_state=self.random_state, n_jobs=self.n_jobs)
        
                    
    
        
    
        
            



import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
import seaborn as sns

from scipy.spatial import distance

def plot_similarity(data, y, sample = True, frac = 1.0, random_state=1, figsize=None):
    """
    data: Dataframe to plot
    y: Column to sort the data by before computing similarity (usually the output)
    sample: (Default = True) Should the data be sampled. Advisable for larger darasets otherewise it takes a long time
    frac = (Default = 1.0 --> Take full dataset) What fraction of the data should be sampled
    """
    sns.set(rc={'image.cmap': 'cubehelix'})
    # get a subset of the data, and normalize it
    if (sample == True):
        df_sub = data.sample(frac=frac, replace=False, random_state=random_state) 
    df_normalized = (df_sub-df_sub.mean())/(df_sub.std())
    df_sorted = df_normalized.sort_values(by=y)
    Y = distance.pdist(df_sorted, "euclidean")
    A = distance.squareform(Y)
    S = 0.5/(1+np.exp(A)) # convert from distance to similarity
    #plot the similarity matrix using seaborn color utilities

    plt.figure(figsize=figsize)
    plt.pcolormesh(S)
    plt.colorbar()


import pandas as pd
import warnings

@pd.api.extensions.register_dataframe_accessor("helper")
class pandas_helper(object):
    """
    # TODO
    # Value Counts df['col_name'].value_counts()
    # NA Values
    # Unique Values (nunique) df['col_names'].nunique()
    # Null Values 
    #   null_vals = df.isnull().sum(axis = 0)
    #   print(null_vals[null_vals != 0])
    # Add options for Datatime, TimeDeltas, etc (check select_dtypes documentation)
    """
    
    def __init__(self, pandas_obj):
        self._validate(pandas_obj)
        self._obj = pandas_obj
        self.cat_features = self._obj.select_dtypes(include=['object','category']).columns
        self.num_features = self._obj.select_dtypes(include=['number']).columns

    @staticmethod
    def _validate(obj):
        """
        Technically not needed for this since helper can only be used with Pandas Dataframe
        Will keep it here for future
        """
        if isinstance(obj, pd.DataFrame):
            pass
        else:
            raise TypeError("Object must be a Pandas Dataframe")
            

    #@property
    
    def describe_categorical(self, verbose=False):
        """
        Prints ategorical variable summaries for a Pandas Dataframe
        Default behavior in pd.DataFrame.describe() is to print only numeric variables if 
        the dataframe contains both numeric and categorical variables. 
        This extension provides more flexibility
        """
        if (self._cat_exists()):
            self.__print_dashes(45)
            print("Summary Statictics for Categorical Variables")
            self.__print_dashes(45)
            print(self._obj[self.cat_features].describe())
            
            if (verbose):
                self.level_counts()
        else:
            warnings.warn("Data does not have any categorical columns")
            
        
    def describe_numeric(self):
        """
        Prints numeric variable summaries for a Pandas Dataframe
        Same as default behavior in pd.DataFrame.describe()
        """
        if (self._num_exists()):
            self.__print_dashes(40)
            print("Summary Statictics for Numeric Variables")
            self.__print_dashes(40)
            print(self._obj[self.num_features].describe())
        else:
            warnings.warn("Data does not have any numeric columns")
    
    def describe(self,verbose=False):
        """
        Prints both numeric and categorical variable summaries for a Pandas Dataframe
        Default behavior in pd.DataFrame.describe() is to print only numeric variables if 
        the dataframe contains both numeric and categorical variables. 
        This extension provides more flexibility
        """
        self.describe_numeric()
        self.describe_categorical(verbose=verbose)
        
    def level_counts(self):
        """
        Printing number of observations in each level of categorical variables
        """
        self.__print_dashes(70)
        print("Printing number of observations in each level of categorical variables")
        self.__print_dashes(70)
        for feature in self.cat_features:
            print(feature)
            print(self._obj[feature].value_counts())
            
    def drop_columns(self, drops, inplace=False):
        """
        Extension of pd.DataFrame.drop(). Instead of giving an error on missing column name,
        this extension only gives a warnign and proceeds with execution
        """
        cols_to_del = []
        cols_not_present = []
        for drop in drops:
            if drop in self._obj.columns:
                cols_to_del.append(drop)
            else:
                cols_not_present.append(drop)
        
        if (len(cols_not_present) > 0):
            warnings.warn("Column(s) {} were not present in the dataframe".format(cols_not_present))
        
        if (inplace):
            self._obj.drop(cols_to_del, axis=1, inplace=inplace)
            self.__set_all_feature_types()
        else:
            return(self._obj.drop(cols_to_del, axis=1, inplace=inplace))
    
    #########################
    #### Private Methods ####
    #########################
    
    def __set_cat_features(self):
        self.cat_features = self._obj.select_dtypes(include=['object','category']).columns
        
    def __set_num_features(self):
        self.num_features = self._obj.select_dtypes(include=['number']).columns
        
    def __set_all_feature_types(self):
        self.__set_cat_features()
        self.__set_num_features()
        
    def _cat_exists(self):
        if (len(self.cat_features) > 0):
            return True
        else:
            return False
            
    def _num_exists(self):
        if (len(self.num_features) > 0):
            return True
        else:
            return False
    
    def __print_dashes(self,num = 20):
        print("-"*num)    
        
    
        


import numpy as np
import pandas as pd
import matplotlib.pyplot as plt

from sklearn import metrics as mt


def print_classification_details(actual,predicted,verbose=False):
    # print the accuracy and confusion matrix 
    cm = mt.confusion_matrix(actual,predicted)
    cr = mt.classification_report(actual,predicted)
    
    print("confusion matrix\n", cm)
    print(cr)
    
    if (verbose == True):
        plot_classification_report(cr)
    
    
def plot_classification_report(cr, title=None, cmap='RdBu'):
    """
    Adapted from https://medium.com/district-data-labs/visual-diagnostics-for-more-informed-machine-learning-7ec92960c96b
    """
    title = title or 'Classification report'
    lines = cr.split('\n')
    classes = []
    matrix = []
  
    for line in lines[2:(len(lines)-5)]:
        s = line.split()
        classes.append(s[0])
        value = [float(x) for x in s[1: len(s) - 1]]
        matrix.append(value)

    fig, ax = plt.subplots(1)

    for column in range(len(matrix)+1):
        for row in range(len(classes)):
            txt = matrix[row][column]
            #ax.text(column,row,matrix[row][column],va='center',ha='center')
            ax.text(column,row,txt,va='center',ha='center',size="x-large",bbox=dict(facecolor='white', alpha=0.5))

    fig = plt.imshow(matrix, interpolation='nearest', cmap=cmap, vmin=0, vmax=1)
    plt.title(title)
    plt.colorbar()
    x_tick_marks = np.arange(len(classes)+1)
    y_tick_marks = np.arange(len(classes))
    plt.xticks(x_tick_marks, ['Precision', 'Recall', 'F1-score'], rotation=45)
    plt.yticks(y_tick_marks, classes)
    plt.ylabel('Classes')
    plt.xlabel('Measures')
    plt.show()


from sklearn.cluster import AgglomerativeClustering
from . import BaseClusterWithN

class AgglomerativeHelper(BaseClusterWithN):
    def __init__(self, X, n_clusters, linkage = "ward", scaled = True, random_state = 101):
        """
        Class to train and evaluate a Agglomerative (Hierarchical) Cluster Model
        """
        super().__init__(X=X, n_clusters=n_clusters, random_state = random_state)
        self.linkage = linkage
        self.cluster_obj = AgglomerativeClustering(n_clusters=self.n_clusters, linkage=self.linkage)
        
        
                    
    
        
    
        
            



import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
import seaborn as sns

import math

from sklearn.model_selection import GridSearchCV


##################################### 
#### Grid Search Hyperparameters ####
#####################################


# Not loaded yet
# If you want to release this, include this in the __init__.py file, and show an example of how to use.

def plot_hyper_validation_curve(estimator, X, y, param_name, param_range
                                ,title = "Validation Curve", xlabel = None, legend_loc='best', logX=False
                                ,cv=None,scoring="accuracy",n_jobs=None,verbose=0, ax = None):
    """
    * Adapted from: 
        - https://chrisalbon.com/machine_learning/model_evaluation/plot_the_validation_curve/
        - https://scikit-learn.org/stable/auto_examples/model_selection/plot_validation_curve.html#sphx-glr-auto-examples-model-selection-plot-validation-curve-py
    
    """
      
    # Calculate accuracy on training and test set using range of parameter values
    train_scores, test_scores = validation_curve(estimator=estimator 
                                                 ,X=X,y=y 
                                                 ,param_name=param_name 
                                                 ,param_range=param_range
                                                 ,cv=cv 
                                                 ,scoring=scoring
                                                 ,verbose=verbose
                                                 ,n_jobs=n_jobs
                                                )
    

    # Calculate mean and standard deviation for training set scores
    train_mean = np.mean(train_scores, axis=1)
    train_std = np.std(train_scores, axis=1)

    # Calculate mean and standard deviation for test set scores
    test_mean = np.mean(test_scores, axis=1)
    test_std = np.std(test_scores, axis=1)
    
    # Clean up for log X axis plotting
    x_label_append = "" # empty if not log scale
    if (logX == True):
        param_range = np.log10(param_range)
        x_label_append = " (log scale)"
    
    # Plot mean accuracy scores for training and test sets
    ax.plot(param_range, train_mean, label="Training Score", color="r")
    ax.plot(param_range, test_mean, label="Cross-Validation Score", color="g")

    # Plot accurancy bands for training and test sets
    ax.fill_between(param_range, train_mean - train_std, train_mean + train_std, color="r",alpha=0.1)
    ax.fill_between(param_range, test_mean - test_std, test_mean + test_std, color="g",alpha=0.1)

    # Add annotations 
    ax.title.set_text(title)
    ax.set_ylabel((scoring + " Score").title())
    if xlabel == None:
        ax.set_xlabel("Hyperparameter: " + param_name + x_label_append)
    else:
        ax.set_xlabel(xlabel + x_label_append)
    
    ax.legend(loc=legend_loc)

def plot_hyper_validation_curves(estimator, X, y, param_grid, scoring
                                  ,logX=False
                                  ,cv=None, refit = 'Accuracy'
                                  ,n_jobs=None,verbose=0,arFigsize=None ):
    
    """
    * Adapted from https://scikit-learn.org/stable/auto_examples/model_selection/plot_multi_metric_evaluation.html
    * Uses GridSearchCV to evaluate many metrics at once. Hence, if evaluating many metrics at once, 
      this is often faster than using validation_curve which will have to be run (trained) once time for each metric)
      
    * Currently supports only one hyperparameter at a time, but can score multiple metrics against that hyperparameter.
    """
    
    gs = GridSearchCV(estimator = estimator
                      ,param_grid=param_grid
                      ,scoring=scoring, cv=cv, refit=refit
                      ,return_train_score=True,n_jobs=-1,verbose=verbose)
    gs.fit(X, y)
    results = gs.cv_results_ 
    
    # Plotting the result
    
    num_rows = math.ceil(len(scoring)/2)
    if (arFigsize == None):
        arFigsize = (12,num_rows*4)
    fig, axes = plt.subplots(num_rows, 2, figsize=arFigsize,squeeze =False)
       
    # For this function, we only expect 1 key
    # If you have more than 1 parameter for GC, then results may not be as expected
    for hyper in param_grid:
        # Get the regular numpy array from the MaskedArray
        X_axis = np.array(results['param_' + hyper].data, dtype=float)    
        
        x_label_append = " "
        if (logX == True):
            X_axis = np.log10(X_axis)
            x_label_append = " (log scale)"

        i = 0
        for scorer in scoring:
            title = "Validation Curve: " + scorer.title()
                       
            ax=axes[math.floor(i/2),i%2]
            for sample, style, color in (('train', 'o-', 'r'), ('test', 'o-', 'g')):
                sample_score_mean = results['mean_%s_%s' % (sample, scorer)]
                sample_score_std = results['std_%s_%s' % (sample, scorer)]

                ax.plot(X_axis, sample_score_mean, style, color=color,
                        #alpha=1 if sample == 'test' else 1,
                        label="%s (%s)" % (scorer, sample))

                ax.fill_between(X_axis, sample_score_mean - sample_score_std,
                                sample_score_mean + sample_score_std,
                                alpha=0.1 if sample == 'test' else 0.1, color=color)

                # Add annotations 
                ax.title.set_text(title)
                ax.set_ylim(-0.1, 1.1)
                ax.set_ylabel(scorer.title() + " Score")
                ax.set_xlabel("Hyperparameter: " + hyper + x_label_append)
                ax.grid()
                ax.legend(loc="best")

            best_index = np.nonzero(results['rank_test_%s' % scorer] == 1)[0][0]
            best_score = results['mean_test_%s' % scorer][best_index]

            # Plot a dotted vertical line at the best score for that scorer marked by x
            ax.plot([X_axis[best_index], ] * 2, [0, best_score],
                    linestyle='-.', color='black', marker='x', markeredgewidth=3, ms=8)

            # Annotate the best score for that scorer
            ax.annotate("%0.2f" % best_score,
                        (X_axis[best_index], best_score + 0.005))


            i = i+1 # Increment for each scorer
    #return(fig)    
    


